/*
 *  aes-glue.c
 *
 *  Written by Jari Ruusu, October 9 2002
 *
 *  Copyright 2001,2002 by Jari Ruusu.
 *  Redistribution of this file is permitted under the GNU Public License.
 */

#include <linux/version.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/errno.h>
#if LINUX_VERSION_CODE >= 0x20500
# include <linux/bio.h>
# include <linux/blkdev.h>
#endif
#if LINUX_VERSION_CODE >= 0x20200
# include <linux/slab.h>
# include <linux/loop.h>
#else
# include <linux/malloc.h>
# include "patched-loop.h"
#endif
#include "aes.h"

#if LINUX_VERSION_CODE >= 0x2052A
typedef sector_t TransferSector_t;
#else
typedef int TransferSector_t;
#endif

int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
          char *loop_buf, int size, TransferSector_t devSect)
{
    register int        x;
    union {
        u_int64_t       h[4];
        u_int32_t       w[8];
        unsigned char   b[32];
    } iv;

    if(!size || (size & 511)) {
        return -EINVAL;
    }
    if(cmd == READ) {
        while(size) {
#if LINUX_VERSION_CODE >= 0x20200
            if(sizeof(TransferSector_t) == 8) {
                iv.h[0] = cpu_to_le64(devSect);
                iv.h[1] = 0;
            } else {
                iv.w[0] = cpu_to_le32(devSect);
                iv.w[3] = iv.w[2] = iv.w[1] = 0;
            }
#else
            iv.b[0] = devSect;
            iv.b[1] = devSect >> 8;
            iv.b[2] = devSect >> 16;
            iv.b[3] = devSect >> 24;
            iv.w[3] = iv.w[2] = iv.w[1] = 0;
#endif
            x = 16;
            do {
                memcpy(&iv.b[16], raw_buf, 16);
                aes_decrypt((aes_context *)lo->key_data, raw_buf, loop_buf);
                *((u_int32_t *)(&loop_buf[ 0])) ^= iv.w[0];
                *((u_int32_t *)(&loop_buf[ 4])) ^= iv.w[1];
                *((u_int32_t *)(&loop_buf[ 8])) ^= iv.w[2];
                *((u_int32_t *)(&loop_buf[12])) ^= iv.w[3];
                raw_buf += 16;
                loop_buf += 16;
                memcpy(&iv.b[0], raw_buf, 16);
                aes_decrypt((aes_context *)lo->key_data, raw_buf, loop_buf);
                *((u_int32_t *)(&loop_buf[ 0])) ^= iv.w[4];
                *((u_int32_t *)(&loop_buf[ 4])) ^= iv.w[5];
                *((u_int32_t *)(&loop_buf[ 8])) ^= iv.w[6];
                *((u_int32_t *)(&loop_buf[12])) ^= iv.w[7];
                raw_buf += 16;
                loop_buf += 16;
#if LINUX_VERSION_CODE >= 0x20500
                cond_resched();
#elif LINUX_VERSION_CODE >= 0x20200
                if(current->need_resched) schedule();
#else
                if(need_resched) schedule();
#endif
            } while(--x);
            size -= 512;
            devSect++;
        }
    } else {
        while(size) {
#if LINUX_VERSION_CODE >= 0x20200
            if(sizeof(TransferSector_t) == 8) {
                iv.h[0] = cpu_to_le64(devSect);
                iv.h[1] = 0;
            } else {
                iv.w[0] = cpu_to_le32(devSect);
                iv.w[3] = iv.w[2] = iv.w[1] = 0;
            }
#else
            iv.b[0] = devSect;
            iv.b[1] = devSect >> 8;
            iv.b[2] = devSect >> 16;
            iv.b[3] = devSect >> 24;
            iv.w[3] = iv.w[2] = iv.w[1] = 0;
#endif
            x = 16;
            do {
                iv.w[0] ^= *((u_int32_t *)(&loop_buf[ 0]));
                iv.w[1] ^= *((u_int32_t *)(&loop_buf[ 4]));
                iv.w[2] ^= *((u_int32_t *)(&loop_buf[ 8]));
                iv.w[3] ^= *((u_int32_t *)(&loop_buf[12]));
                aes_encrypt((aes_context *)lo->key_data, &iv.b[0], raw_buf);
                memcpy(&iv.b[0], raw_buf, 16);
                loop_buf += 16;
                raw_buf += 16;
                iv.w[0] ^= *((u_int32_t *)(&loop_buf[ 0]));
                iv.w[1] ^= *((u_int32_t *)(&loop_buf[ 4]));
                iv.w[2] ^= *((u_int32_t *)(&loop_buf[ 8]));
                iv.w[3] ^= *((u_int32_t *)(&loop_buf[12]));
                aes_encrypt((aes_context *)lo->key_data, &iv.b[0], raw_buf);
                memcpy(&iv.b[0], raw_buf, 16);
                loop_buf += 16;
                raw_buf += 16;
#if LINUX_VERSION_CODE >= 0x20500
                cond_resched();
#elif LINUX_VERSION_CODE >= 0x20200
                if(current->need_resched) schedule();
#else
                if(need_resched) schedule();
#endif
            } while(--x);
            size -= 512;
            devSect++;
        }
    }
    return(0);
}

int keySetup_aes(struct loop_device *lo, struct loop_info *info)
{
    lo->key_data = (aes_context *) kmalloc(sizeof(aes_context), GFP_KERNEL);
    if(!lo->key_data) return(-ENOMEM);

    aes_set_key((aes_context *)lo->key_data, &info->lo_encrypt_key[0],
            info->lo_encrypt_key_size, 0);
    memset(&info->lo_encrypt_key[0], 0, sizeof(info->lo_encrypt_key));
    return(0);
}

int keyClean_aes(struct loop_device *lo)
{
    if(lo->key_data) {
        memset(lo->key_data, 0, sizeof(aes_context));
        kfree(lo->key_data);
        lo->key_data = 0;
    }
    return(0);
}

#if LINUX_VERSION_CODE >= 0x20200

static struct loop_func_table funcs_aes = { 
    number:     16,     /* 16 == AES */
    transfer:   transfer_aes,
    init:       keySetup_aes,
    release:    keyClean_aes
};

int init_module_aes(void)
{
    if(loop_register_transfer(&funcs_aes)) {
        printk("loop: unable to register AES transfer\n");
        return -EIO;
    }
    return 0;
}

void cleanup_module_aes(void)
{
    if(loop_unregister_transfer(funcs_aes.number)) {
        printk("loop: unable to unregister AES transfer\n");
    }
}

#endif
